% Makefile metrics analysis
% Jim Cordy, Ecole Polytechnique, April 2014
% Douglas Martin, Queen's University, January 2017

% Version 2.0, 20 Jan 2017

include "make.grm"

redefine Rule
	[space?] [Targets] [ColonOrDoubleColon] [Patterns] [TargetAssignments] [Dependencies] [Recipe]
end define

redefine Assignment
	[WS] [PrivateExportOverride?] [WS] [VarName] [WS] [AssignmentOp] [WS] [BSWS?] [Expr?] %% [EOL?]
		[AssignmentContinuation*] [EOL?]
end redefine

define VarName
		[VpathId]
	|	[Id]
end define

define VpathId
	'VPATH
end define

redefine Target
	[SpecialTarget] [WS]
    |	[SuffixTarget] [WS]
    |	...
end redefine

define IdPart
		[SpecialTarget] [WS]
	|	...
end define

define SpecialTarget
		'.PHONY | '.phony
	|	'.SUFFIXES | '.suffixes
	|	'.DEFAULT | '.default
    |	'.PRECIOUS | '.precious   
	|	'.INTERMEDIATE | '.intermediate
    |	'.SECONDARY | '.secondary
    |	'.SECONDEXPANSION | '.secondexpansion
    |	'.DELETE_ON_ERROR | '.delete_on_error
    |	[IgnoreTarget]
	|	'.LOW_RESOLUTION_TIME | '.low_resolution_time
	|	[SilentTarget]
	|	'.EXPORT_ALL_VARIABLES | '.export_all_variables
	|	'.NOTPARALLEL | '.notparallel
	|	'.ONESHELL | '.oneshell
	|	'.POSIX | '.posix
    |	'.NO_EXPORT | '.no_export
    |	'.MAKE | '.make
end define

define SilentTarget
	'.SILENT | '.silent
end define

define IgnoreTarget
	'.IGNORE | '.ignore
end define

define SuffixTarget
	'. [id]
end define


define MakeTerm
	'/bin/make | 'make | '$(make) | '$(MAKE) | '$(GMAKE)
end redefine

define AutomakeTerm
	'$(automake) | '$(AUTOMAKE)
end redefine

define CMakeTerm
	'$(CMAKE_COMMAND)
end redefine

define QMakeTerm
	'$(QMAKE)
end redefine

redefine AtMinusPlus
	[At] | [Minus] | [Plus]
end redefine

define At
	'@ 
end define

define Minus
	'- 
end define

define Plus
	'+ 
end define

redefine AssignmentOp
	[LazyAssignmentOp]
    |	[StrictAssignmentOp]
    |	[IterativeAssignmentOp]
    |	[ConditionalAssignmentOp]
end redefine

define LazyAssignmentOp
	'=
end define

define StrictAssignmentOp
	':=
end define

define IterativeAssignmentOp
	'+=
end define

define ConditionalAssignmentOp
	'?= 
end define

define ColonOrDoubleColon
	[DoubleColon] | [Colon]
end define

define DoubleColon
	'::
end define

define Colon
	':
end define

compounds
	$(@D) $(@F) $(<D) $(<F) $(?D) $(?F) $(^D) $(^F) $(+D) $(+F) $(*D) $(*F) $(%D) $(%F)
end compounds

redefine AutoEval
		[AE_At]
	|	[AE_LT]
	|	[AE_QM]
	|	[AE_Chev]
	|	[AE_Plus]
	|	[AE_Star]
	|	[AE_Perc]
	|	[AE_Pipe]
	|	[AE_Obsolete]
end redefine

define AE_At
	'$@
end define

define AE_LT
	'$<
end define

define AE_QM
	'$?
end define

define AE_Chev
	'$^
end define

define AE_Plus
	'$+
end define

define AE_Star
	'$*
end define

define AE_Perc
	'$%
end define

define AE_Pipe
	'$|
end define

define AE_Obsolete
	'$(@D) | '$(@F) | '$(<D) | '$(<F) | '$(?D) | '$(?F) | '$(^D) | '$(^F) | '$(+D) | '$(+F) | '$(*D) | '$(*F) | '$(%D) | '$(%F)
end define

redefine RecipeTerm
		[MakeTerm]
    |	[AutomakeTerm]
    |	[CMakeTerm]
	|	[QMakeTerm]
    |	[IfRecipe] 
    |	[AutoEval] 
	| 	[ChangeDir] | [PushPopDir]
    |	[Id]			% includes Evaluation and FunctionCall
    |	[not Id] [Literal] 
    |	[BSWS]			% empty line?
end redefine

define ChangeDir
	'cd | 'chdir | '$(CD) | '$(CHDIR) | '${CD} | '${CHDIR} | '$(cd) | '$(chdir) | '${cd} | '${chdir}
end define
	
define PushPopDir
	'pushd | 'popd  
end define	
	
function main
    match [program] Makefile [program]

    % Name of makefile
    import TXLinput [stringlit]
    construct _ [stringlit]
	TXLinput [putp "Makefile: %"]

    % Size of whole makefile
    construct EOLs [newline*]
	_ [^ Makefile]
    construct _ [number]
	_ [length EOLs] [putp "lines: %"]

    % Continuations
    construct Backslashes [BSWS*]
	_ [^ Makefile]
    construct _ [number]
 	_ [length Backslashes] [putp "continuations: %"]

    % Comments
    construct Comments [comment*]
	_ [^ Makefile]
    construct _ [number]
	_ [length Comments] [putp "comments: %"]

	% Generated?	
	construct Generator [stringlit]
	_ [isAutotools Makefile] [isCMake Makefile] [isQMake Makefile] 
	%_ [isGenerated Makefile] [isAutotools Makefile] [isCMake Makefile] [isQMake Makefile] [isIMake Makefile]	
	construct _ [stringlit]
	_ [+ Generator] [putp "generated: %"]
	
	% cds
	construct Cds [ChangeDir*]
	_ [^ Makefile] 
	construct _ [number]
	_ [length Cds] [putp "directory changes (cd): %"]
	
    % Macros
    construct Macros [Definition*]
	_ [^ Makefile]
    construct _ [number]
	_ [length Macros] [putp "macros: %"]

    % Vpaths
    construct Vpaths [Vpath*]
	_ [^ Makefile]
    construct _ [number]
 	_ [length Vpaths] [putp "vpath directives: %"]

    % Includes
    construct Includes [Include*]
	_ [^ Makefile]
    construct _ [number]
 	_ [length Includes] [putp "includes: %"]

    % Ifdef / Ifeq
    construct Ifs [IfStatement*]
	_ [^ Makefile]
    construct _ [number]
 	_ [length Ifs] [putp "ifdef/ifeq directives: %"]

    % Assignments
    construct Assigns [Assignment*]
	_ [^ Makefile]
    construct _ [number]
	_ [length Assigns] [putp "assignments: %"]
	
	construct VpathVars [VpathId*]
	_ [^ Assigns]
    construct _ [number]
 	_ [length VpathVars] [putp "VPATH variable assignments: %"]

    % Assignment variable evaluations
    construct AssignEvals [Evaluation*]
	_ [^ Assigns]
    construct NAssignEvals [number]
 	_ [length AssignEvals] [putp "assignment variable references: %"]
	
    construct UniqueAssignEvals [Evaluation*]
	_ [getUniqueAssignmentVarRefs each Assigns]
    construct NUniqueAssignEvals [number]
 	_ [length UniqueAssignEvals] [putp "unique assignment variable references: %"]
		
    % Assignment function calls
    construct AssignFuncs [FunctionCall*]
	_ [^ Assigns]
    construct NAssignFuncs [number]
 	_ [length AssignFuncs] [putp "assignment function calls: %"]

    % Assignment auto evaluations	$@ $< $? $^ $+ $*	
	construct AssignAE_At [AE_At*]
	_ [^ Assigns]
	construct _ [number]
 	_ [length AssignAE_At] [putp "assignment $@ references: %"]
	
	construct AssignAE_LT [AE_LT*]
	_ [^ Assigns]
	construct _ [number]
 	_ [length AssignAE_LT] [putp "assignment $< references: %"]
	
	construct AssignAE_QM [AE_QM*]
	_ [^ Assigns]
	construct _ [number]
 	_ [length AssignAE_QM] [putp "assignment $? references: %"]
	
	construct AssignAE_Chev [AE_Chev*]
	_ [^ Assigns]
	construct _ [number]
 	_ [length AssignAE_Chev] [putp "assignment $^ references: %"]
	
	construct AssignAE_Plus [AE_Plus*]
	_ [^ Assigns]
	construct _ [number]
 	_ [length AssignAE_Plus] [putp "assignment $+ references: %"]
	
	construct AssignAE_Star [AE_Star*]
	_ [^ Assigns]
	construct _ [number]
 	_ [length AssignAE_Star] [putp "assignment $* references: %"]
	
	construct AssignAE_Perc [AE_Perc*]
	_ [^ Assigns]
	construct AssignAE_Perc_Length [number]
 	_ [length AssignAE_Perc] 
	construct AssignAE_Perc_Length_String [stringlit]
	_ [unparse AssignAE_Perc_Length]
	construct AssignAE_Perc_String1 [stringlit]
	"assignment $% references: " 
	construct AssignAE_Perc_String2 [stringlit]
	AssignAE_Perc_String1 [+ AssignAE_Perc_Length_String]
	construct _ [stringlit]
	AssignAE_Perc_String2 [put]
	
	construct AssignAE_Pipe [AE_Pipe*]
	_ [^ Assigns]
	construct _ [number]
 	_ [length AssignAE_Pipe] [putp "assignment $| references: %"]
	
	construct AssignAE_Obsolete [AE_Obsolete*]
	_ [^ Assigns]
    construct _ [number]
 	_ [length AssignAE_Obsolete] [putp "assignment total obsolete autoeval references: %"]
	
	construct AssignAutoEvals [AutoEval*]
	_ [^ Assigns]
    construct _ [number]
 	_ [length AssignAutoEvals] [putp "assignment total autoeval references: %"]

    % Lazy assignments =
    construct LazyAssignments [LazyAssignmentOp*]
	_ [^ Assigns]
    construct _ [number]
 	_ [length LazyAssignments] [putp "assignment lazy assignments: %"]

    % Strict assignments :=
    construct StrictAssignments [StrictAssignmentOp*]
	_ [^ Assigns]
    construct _ [number]
 	_ [length StrictAssignments] [putp "assignment strict assignments: %"]

    % Iterative assignments +=
    construct IterativeAssignments [IterativeAssignmentOp*]
	_ [^ Assigns]
    construct _ [number]
 	_ [length IterativeAssignments] [putp "assignment iterative assignments: %"]

    % Conditional assignments ?=
    construct ConditionalAssignments [ConditionalAssignmentOp*]
	_ [^ Assigns]
    construct _ [number]
 	_ [length ConditionalAssignments] [putp "assignment conditional assignments: %"]
	
	construct AssignPaths [Assignment*]
	_ [findPaths_Assign each Assigns]
	construct NAssignPaths [number]
	_ [length AssignPaths] [putp "assignment paths: %"]

    % Rules
    construct Rules [Rule*]
	_ [^ Makefile]
    construct _ [number]
	_ [length Rules] [putp "rules (all): %"]
	
	% Rules
    construct SingleColonRules [Colon*]
	_ [^ Makefile]
    construct _ [number]
	_ [length SingleColonRules] [putp "rules (single colon): %"]
	
	% Rules
    construct DoubleColonRules [DoubleColon*]
	_ [^ Makefile]
    construct _ [number]
	_ [length DoubleColonRules] [putp "rules (double colon): %"]

	construct PatternRules [Rule*]
	_ [getPatternRules each Rules]
	construct _ [number]
	_ [length PatternRules] [putp "pattern rules: %"]

	construct StaticPatternRules [Pattern*]
	_ [^ Rules]
	construct _ [number]
	_ [length StaticPatternRules] [putp "static pattern rules: %"]
	
    % Targets 
    construct Targets [Target*]
	_ [^ Makefile]
    construct _ [number]
 	_ [length Targets] [putp "rule targets: %"]

    % Special targets .PHONY .INTERMEDIATE .SECONDARY .PRECIOUS .DELETE_ON_ERROR
    construct SpecialTargets [SpecialTarget*]
	_ [^ Targets]
    construct _ [number]
 	_ [length SpecialTargets] [putp "rule target special targets: %"]
	
	construct SilentTargets [SilentTarget*]
	_ [^ SpecialTargets]
    construct _ [number]
 	_ [length SilentTargets] [putp "rule target .SILENT targets: %"]
	
	construct IgnoreTargets [IgnoreTarget*]
	_ [^ SpecialTargets]
    construct _ [number]
 	_ [length IgnoreTargets] [putp "rule target .IGNORE targets: %"]

    % Suffix targets .c.o
    construct SuffixTargets [SuffixTarget*]
	_ [^ Targets]
    construct _ [number]
 	_ [length SuffixTargets] [putp "rule target suffix targets: %"]

	construct TargetAssigns [TargetAssignments*]
	_ [^ Rules]
	construct _ [number]
	_ [length TargetAssigns] [putp "target-specific assignments: %"]
	
    % Target variable evaluations
    construct TargetEvals [Evaluation*]
	_ [^ Targets]
    construct NTargetEvals [number]
 	_ [length TargetEvals] [putp "rule target variable references: %"]

	construct UniqueTargetVarRefs [Evaluation*]
	_ [getUniqueTargetVarRefs each Rules]
	construct _ [number]
	_ [length UniqueTargetVarRefs] [putp "unique target variable references: %"]
	
    % Target function calls
    construct TargetFuncs [FunctionCall*]
	_ [^ Targets]
    construct NTargetFuncs [number]
 	_ [length TargetFuncs] [putp "rule target function calls: %"]

	construct TargetPaths [Target*]
	_ [findPaths_Target each Targets]
	construct NTargetPaths [number]
	_ [length TargetPaths] [putp "rule target paths: %"]
	
    % Dependencies
    construct Dependencies [Dependency*]
	_ [^ Makefile]
    construct _ [number]
 	_ [length Dependencies] [putp "rule dependencies: %"]

    % Dependency variable evaluations
    construct DependencyEvals [Evaluation*]
	_ [^ Dependencies]
    construct NDependencyEvals [number]
	_ [length DependencyEvals] [putp "rule dependency variable references: %"]

	construct UniqueDependencyVarRefs [Evaluation*]
	_ [getUniqueDependencyVarRefs each Rules]
	construct _ [number]
	_ [length UniqueDependencyVarRefs] [putp "unique dependency variable references: %"]
	
    % Dependency Function calls
    construct DependencyFuncs [FunctionCall*]
	_ [^ Dependencies]
    construct NDependencyFuncs [number]
 	_ [length DependencyFuncs] [putp "rule dependency function calls: %"]

    % Dependency Auto evaluations	$@ $< $? $^ $+ $*
	construct DependencyAE_At [AE_At*]
	_ [^ Dependencies]
	construct _ [number]
 	_ [length DependencyAE_At] [putp "rule dependency $@ references: %"]
	
	construct DependencyAE_LT [AE_LT*]
	_ [^ Dependencies]
	construct _ [number]
 	_ [length DependencyAE_LT] [putp "rule dependency $< references: %"]
	
	construct DependencyAE_QM [AE_QM*]
	_ [^ Dependencies]
	construct _ [number]
 	_ [length DependencyAE_QM] [putp "rule dependency $? references: %"]
	
	construct DependencyAE_Chev [AE_Chev*]
	_ [^ Dependencies]
	construct _ [number]
 	_ [length DependencyAE_Chev] [putp "rule dependency $^ references: %"]
	
	construct DependencyAE_Plus [AE_Plus*]
	_ [^ Dependencies]
	construct _ [number]
 	_ [length DependencyAE_Plus] [putp "rule dependency $+ references: %"]
	
	construct DependencyAE_Star [AE_Star*]
	_ [^ Dependencies]
	construct _ [number]
 	_ [length DependencyAE_Star] [putp "rule dependency $* references: %"]
	
	construct DependencyAE_Perc [AE_Perc*]
	_ [^ Dependencies]
	construct DependencyAE_Perc_Length [number]
 	_ [length DependencyAE_Perc] 
	construct DependencyAE_Perc_Length_String [stringlit]
	_ [unparse DependencyAE_Perc_Length]
	construct DependencyAE_Perc_String1 [stringlit]
	"rule dependency $% references: " 
	construct DependencyAE_Perc_String2 [stringlit]
	DependencyAE_Perc_String1 [+ DependencyAE_Perc_Length_String]
	construct _ [stringlit]
	DependencyAE_Perc_String2 [put]	
	
	construct DependencyAE_Pipe [AE_Pipe*]
	_ [^ Dependencies]
	construct _ [number]
 	_ [length DependencyAE_Pipe] [putp "rule dependency $| references: %"]

	construct DependencyAE_Obsolete [AE_Obsolete*]
	_ [^ Dependencies]
    construct _ [number]
 	_ [length DependencyAE_Obsolete] [putp "rule dependency total obsolete autoeval references: %"]
	
    construct DependencyAutoEvals [AutoEval*]
	_ [^ Dependencies]
    construct _ [number]
 	_ [length DependencyAutoEvals] [putp "rule dependency total autoeval references: %"]

	construct DependencyPaths [Dependency*]
	_ [findPaths_Dependency each Dependencies]
	construct NDependencyPaths [number]
	_ [length DependencyPaths] [putp "rule dependency paths: %"]
	
    % Recipe commands
    construct RecipeCommands [RecipeExpr*]
	_ [^ Makefile]
    construct _ [number]
 	_ [length RecipeCommands] [putp "rule recipe commands: %"]

    % @ commands
    construct Ats [At*]
	_ [^ Rules]
    construct _ [number]
 	_ [length Ats] [putp "rule @ recipe commands: %"]

    % - commands
    construct Minuses [Minus*]
	_ [^ Rules]
    construct _ [number]
 	_ [length Minuses] [putp "rule - recipe commands: %"]
	
	% - commands
    construct Pluses [Plus*]
	_ [^ Rules]
    construct _ [number]
 	_ [length Pluses] [putp "rule + recipe commands: %"]

    % Recursive make
    construct Makes [MakeTerm*]
	_ [^ RecipeCommands]
    construct _ [number]
 	_ [length Makes] [putp "rule recipe recursive makes: %"]

    % Recursive automake
    construct Automakes [AutomakeTerm*]
	_ [^ RecipeCommands]
    construct _ [number]
 	_ [length Automakes] [putp "rule recipe recursive automakes: %"]

	% Recursive CMake
    construct CMakes [CMakeTerm*]
	_ [^ RecipeCommands]
    construct _ [number]
 	_ [length CMakes] [putp "rule recipe recursive cmakes: %"]
	
	% Recursive QMake
    construct QMakes [QMakeTerm*]
	_ [^ RecipeCommands]
    construct _ [number]
 	_ [length QMakes] [putp "rule recipe recursive qmakes: %"]
	
    % Recipe variable evaluations
    construct RecipeEvals [Evaluation*]
	_ [^ RecipeCommands]
    construct NRecipeEvals [number]
 	_ [length RecipeEvals] [putp "rule recipe variable references: %"]

	construct UniqueRecipeVarRefs [Evaluation*]
	_ [getUniqueRecipeVarRefs each Rules]
	construct _ [number]
	_ [length UniqueRecipeVarRefs] [putp "unique recipe variable references: %"]
	
    % Recipe function calls
    construct RecipeFuncs [FunctionCall*]
	_ [^ RecipeCommands]
    construct NRecipeFuncs [number]
 	_ [length RecipeFuncs] [putp "rule recipe function calls: %"]

    % Recipe auto evaluations	$@ $< $? $^ $+ $*
	construct RecipeAE_At [AE_At*]
	_ [^ RecipeCommands]
	construct _ [number]
 	_ [length RecipeAE_At] [putp "rule recipe $@ references: %"]
	
	construct RecipeAE_LT [AE_LT*]
	_ [^ RecipeCommands]
	construct _ [number]
 	_ [length RecipeAE_LT] [putp "rule recipe $< references: %"]
	
	construct RecipeAE_QM [AE_QM*]
	_ [^ RecipeCommands]
	construct _ [number]
 	_ [length RecipeAE_QM] [putp "rule recipe $? references: %"]
	
	construct RecipeAE_Chev [AE_Chev*]
	_ [^ RecipeCommands]
	construct _ [number]
 	_ [length RecipeAE_Chev] [putp "rule recipe $^ references: %"]
	
	construct RecipeAE_Plus [AE_Plus*]
	_ [^ RecipeCommands]
	construct _ [number]
 	_ [length RecipeAE_Plus] [putp "rule recipe $+ references: %"]
	
	construct RecipeAE_Star [AE_Star*]
	_ [^ RecipeCommands]
	construct _ [number]
 	_ [length RecipeAE_Star] [putp "rule recipe $* references: %"]
	
	construct RecipeAE_Perc [AE_Perc*]
	_ [^ RecipeCommands]
	construct RecipeAE_Perc_Length [number]
 	_ [length RecipeAE_Perc] 
	construct RecipeAE_Perc_Length_String [stringlit]
	_ [unparse RecipeAE_Perc_Length]
	construct RecipeAE_Perc_String1 [stringlit]
	"rule recipe $% references: " 
	construct RecipeAE_Perc_String2 [stringlit]
	RecipeAE_Perc_String1 [+ RecipeAE_Perc_Length_String]
	construct _ [stringlit]
	RecipeAE_Perc_String2 [put]
	
	construct RecipeAE_Pipe [AE_Pipe*]
	_ [^ RecipeCommands]
	construct _ [number]
 	_ [length RecipeAE_Pipe] [putp "rule recipe $| references: %"]
	
	construct RecipeAE_Obsolete [AE_Obsolete*]
	_ [^ RecipeCommands]
    construct _ [number]
 	_ [length RecipeAE_Obsolete] [putp "rule recipe total obsolete autoeval references: %"]
	
    construct RecipeAutoEvals [AutoEval*]
	_ [^ RecipeCommands]
    construct _ [number]
 	_ [length RecipeAutoEvals] [putp "rule recipe total autoeval references: %"]
	
	construct RecipePaths [RecipeExpr*]
	_ [findPaths_Recipe each RecipeCommands]
	construct NRecipePaths [number]
	_ [length RecipePaths] [putp "rule recipe paths: %"]
	
	
	%total variable references	
	construct AllVarRefs [Evaluation*]
	_ [^ Makefile]
	construct _ [number]
 	_ [length AllVarRefs] [putp "total variable references: %"]
	
	%total function calls
	construct AllFuncCalls [FunctionCall*]
	_ [^ Makefile]
	construct _ [number]
 	_ [length AllFuncCalls] [putp "total function calls: %"]
	
	
	%construct RecursiveFuncCalls [FunctionCall*]
	%_ [^ AllFuncCalls]
	export TotalEmbeddedFunctions [number]
		0
	construct _ [program]
		Makefile [countFunctionCalls]
	construct _ [program]
		Makefile [countErrorFunctionCalls]
	import TotalEmbeddedFunctions
	construct _ [number]
 	TotalEmbeddedFunctions [putp "total embedded function calls: %"]
	
	construct AllVars [Evaluation*]
	_ [^ Makefile]
	construct UniqueVars [Evaluation*]
	_ [getUniqueVars each AllVars]
	construct _ [number]
 	_ [length UniqueVars] [putp "total unique variable references: %"]
	
	construct AllSubstFuncs [FunctionCall*]
	_ [getFunctions 'subst each AllFuncCalls]
	construct _ [number]
 	_ [length AllSubstFuncs] [putp "total subst function calls: %"]
	
	construct AllPatSubstFuncs [FunctionCall*]
	_ [getFunctions 'patsubst each AllFuncCalls]
	construct _ [number]
 	_ [length AllPatSubstFuncs] [putp "total patsubst function calls: %"]

	construct AllStripFuncs [FunctionCall*]
	_ [getFunctions 'strip each AllFuncCalls]
	construct _ [number]
 	_ [length AllStripFuncs] [putp "total strip function calls: %"]
	
	construct AllFindStringFuncs [FunctionCall*]
	_ [getFunctions 'findstring each AllFuncCalls]
	construct _ [number]
 	_ [length AllFindStringFuncs] [putp "total findstring function calls: %"]
	
	construct AllFilterFuncs [FunctionCall*]
	_ [getFunctions 'filter each AllFuncCalls]
	construct _ [number]
 	_ [length AllFilterFuncs] [putp "total filter function calls: %"]
	
	construct AllFilterOutFuncs [FunctionCall*]
	_ [getFunctions 'filter-out each AllFuncCalls]
	construct _ [number]
 	_ [length AllFilterOutFuncs] [putp "total filter-out function calls: %"]
	
	construct AllSortFuncs [FunctionCall*]
	_ [getFunctions 'sort each AllFuncCalls]
	construct _ [number]
 	_ [length AllSortFuncs] [putp "total sort function calls: %"]
	
	construct AllWordFuncs [FunctionCall*]
	_ [getFunctions 'word each AllFuncCalls]
	construct _ [number]
 	_ [length AllWordFuncs] [putp "total word function calls: %"]
	
	construct AllWordListFuncs [FunctionCall*]
	_ [getFunctions 'wordlist each AllFuncCalls]
	construct _ [number]
 	_ [length AllWordListFuncs] [putp "total wordlist function calls: %"]
	
	construct AllWordsFuncs [FunctionCall*]
	_ [getFunctions 'words each AllFuncCalls]
	construct _ [number]
 	_ [length AllWordsFuncs] [putp "total words function calls: %"]
	
	construct AllFirstWordFuncs [FunctionCall*]
	_ [getFunctions 'firstword each AllFuncCalls]
	construct _ [number]
 	_ [length AllFirstWordFuncs] [putp "total firstword function calls: %"]
	
	construct AllLastWordFuncs [FunctionCall*]
	_ [getFunctions 'lastword each AllFuncCalls]
	construct _ [number]
 	_ [length AllLastWordFuncs] [putp "total lastword function calls: %"]
	
	construct AllDirFuncs [FunctionCall*]
	_ [getFunctions 'dir each AllFuncCalls]
	construct _ [number]
 	_ [length AllDirFuncs] [putp "total dir function calls: %"]
	
	construct AllNotDirFuncs [FunctionCall*]
	_ [getFunctions 'notdir each AllFuncCalls]
	construct _ [number]
 	_ [length AllNotDirFuncs] [putp "total notdir function calls: %"]
	
	construct AllSuffixFuncs [FunctionCall*]
	_ [getFunctions 'suffix each AllFuncCalls]
	construct _ [number]
 	_ [length AllSuffixFuncs] [putp "total suffix function calls: %"]
	
	construct AllBasenameFuncs [FunctionCall*]
	_ [getFunctions 'basename each AllFuncCalls]
	construct _ [number]
 	_ [length AllBasenameFuncs] [putp "total basename function calls: %"]
	
	construct AllAddSuffixFuncs [FunctionCall*]
	_ [getFunctions 'addsuffix each AllFuncCalls]
	construct _ [number]
 	_ [length AllAddSuffixFuncs] [putp "total addsuffix function calls: %"]
	
	construct AllAddPrefixFuncs [FunctionCall*]
	_ [getFunctions 'addprefix each AllFuncCalls]
	construct _ [number]
 	_ [length AllAddPrefixFuncs] [putp "total addprefix function calls: %"]
	
	construct AllJoinFuncs [FunctionCall*]
	_ [getFunctions 'join each AllFuncCalls]
	construct _ [number]
 	_ [length AllJoinFuncs] [putp "total join function calls: %"]
	
	construct AllWildcardFuncs [FunctionCall*]
	_ [getFunctions 'wildcard each AllFuncCalls]
	construct _ [number]
 	_ [length AllWildcardFuncs] [putp "total wildcard function calls: %"]
	
	construct AllRealPathFuncs [FunctionCall*]
	_ [getFunctions 'realpath each AllFuncCalls]
	construct _ [number]
 	_ [length AllRealPathFuncs] [putp "total realpath function calls: %"]
	
	construct AllAbsPathFuncs [FunctionCall*]
	_ [getFunctions 'abspath each  AllFuncCalls]
	construct _ [number]
 	_ [length AllAbsPathFuncs] [putp "total abspath function calls: %"]
	
	construct AllErrorFuncs [FunctionCall*]
	_ [getErrorFunctions each  AllFuncCalls]
	construct _ [number]
 	_ [length AllErrorFuncs] [putp "total error function calls: %"]
	
	construct AllWarningFuncs [FunctionCall*]
	_ [getWarningFunctions each  AllFuncCalls]
	construct _ [number]
 	_ [length AllWarningFuncs] [putp "total warning function calls: %"]
	
	construct AllInfoFuncs [FunctionCall*]
	_ [getInfoFunctions each  AllFuncCalls]
	construct _ [number]
 	_ [length AllInfoFuncs] [putp "total info function calls: %"]
	
	construct AllFileFuncs [FunctionCall*]
	_ [getFunctions 'file each  AllFuncCalls]
	construct _ [number]
 	_ [length AllFileFuncs] [putp "total file function calls: %"]
	
	construct AllCallFuncs [FunctionCall*]
	_ [getFunctions 'call each  AllFuncCalls]
	construct _ [number]
 	_ [length AllCallFuncs] [putp "total call function calls: %"]
	
	construct AllValueFuncs [FunctionCall*]
	_ [getFunctions 'value each  AllFuncCalls]
	construct _ [number]
 	_ [length AllValueFuncs] [putp "total value function calls: %"]
	
	construct AllEvalFuncs [FunctionCall*]
	_ [getFunctions 'eval each  AllFuncCalls]
	construct _ [number]
 	_ [length AllEvalFuncs] [putp "total eval function calls: %"]
	
	construct AllOriginFuncs [FunctionCall*]
	_ [getFunctions 'origin each  AllFuncCalls]
	construct _ [number]
 	_ [length AllOriginFuncs] [putp "total origin function calls: %"]
	
	construct AllFlavorFuncs [FunctionCall*]
	_ [getFunctions 'flavor each  AllFuncCalls]
	construct _ [number]
 	_ [length AllFlavorFuncs] [putp "total flavor function calls: %"]
	
	construct AllShellFuncs [FunctionCall*]
	_ [getShellFunctions each  AllFuncCalls]
	construct _ [number]
 	_ [length AllShellFuncs] [putp "total shell function calls: %"]
	
	construct AllGuileFuncs [FunctionCall*]
	_ [getFunctions 'guile each  AllFuncCalls]
	construct _ [number]
 	_ [length AllGuileFuncs] [putp "total guile function calls: %"]
	
	construct AllForeachFuncs [FunctionCall*]
	_ [getFunctions 'foreach each  AllFuncCalls]
	construct _ [number]
 	_ [length AllForeachFuncs] [putp "total foreach function calls: %"]
	
	construct AllIfFuncs [FunctionCall*]
	_ [getFunctions 'if each  AllFuncCalls]
	construct _ [number]
 	_ [length AllIfFuncs] [putp "total if function calls: %"]
	
	construct AllOrFuncs [FunctionCall*]
	_ [getFunctions 'or each  AllFuncCalls]
	construct _ [number]
 	_ [length AllOrFuncs] [putp "total or function calls: %"]
	
	construct AllAndFuncs [FunctionCall*]
	_ [getFunctions 'and each  AllFuncCalls]
	construct _ [number]
 	_ [length AllAndFuncs] [putp "total and function calls: %"]
	
	construct AllIds [id*]
	_ [^ Makefile]
	construct IdPaths [id*]
	_ [findPaths_Id each AllIds]
	construct NIdPaths [number]
	_ [length IdPaths] [putp "total paths: %"]
end function

function isGenerated Makefile [program]
	replace [stringlit]
		_ [stringlit]
	where
		Makefile [containsComment "generated"] [containsComment "automatically"]
	by 
		"other?"
end function 

function isAutotools Makefile [program]
	replace [stringlit]
		_ [stringlit]
	where
		Makefile [containsComment "generated automatically by automake"] [containsComment "derived from Makefile"] [containsComment "Makefile.am"] %[containsComment "by configure"]
	by 
		"autotools"
end function 

function isCMake Makefile [program]
	replace [stringlit]
		_ [stringlit]
	where
		Makefile [containsComment "CMAKE generated file"]
	by 
		"cmake"
end function 

function isQMake Makefile [program]
	replace [stringlit]
		_ [stringlit]
	where
		Makefile [containsComment "Generated by qmake"]
	by 
		"qmake"
end function 

function isIMake Makefile [program]
	replace [stringlit]
		_ [stringlit]	
	where
		Makefile [containsComment "generated by imake"]
	by 
		"imake"
end function 

function containsComment X [stringlit] 
	match * [comment]
		C [comment]
	where 
		C [grep X]
end function

function getPatternRules R [Rule]
	replace [Rule*]
		PrevRules [Rule*]
	construct Targets [Target*]
		_ [^ R]
	where	
		Targets [containsPerc]
	where not
		Targets [containsPercEquals]
	by
		PrevRules [. R]
end function

function containsPerc
	match * [Target]
		T [Target]
	construct TargetStr [stringlit]
		_ [unparse T]
	where	
		TargetStr [grep "%"]
end function

function containsPercEquals
	match * [Target]
		T [Target]
	construct TargetStr [stringlit]
		_ [unparse T]
	where	
		TargetStr [grep "%="]
end function

function getFunctions FuncName [id] Func [FunctionCall]
	replace [FunctionCall*]
		PrevFuncs [FunctionCall*]
	deconstruct Func
		%%%'$ '( FuncName _ [tabspace] _ [list WSExpr+] _ [WS] ')
		'$ '( FuncName _ [tabspace] _  [list FunctionArg+] _ [WS] ') _ [WS]	
	by
		PrevFuncs [. Func]
end function


function getErrorFunctions Func [FunctionCall]
	replace [FunctionCall*]
		PrevFuncs [FunctionCall*]
	deconstruct Func
		'$ '( X [ErrorName] _ [tabspace] _ [ErrorExpr*] _ [WS] ') _ [WS]
	deconstruct X
		'error _ [': ?]
	by
		PrevFuncs [. Func]
end function

function getWarningFunctions Func [FunctionCall]
	replace [FunctionCall*]
		PrevFuncs [FunctionCall*]
	deconstruct Func
		'$ '( X [ErrorName] _ [tabspace] _ [ErrorExpr*] _ [WS] ') _ [WS]
	deconstruct X
		'warning _ [': ?]
	by
		PrevFuncs [. Func]
end function

function getInfoFunctions Func [FunctionCall]
	replace [FunctionCall*]
		PrevFuncs [FunctionCall*]
	deconstruct Func
		'$ '( X [ErrorName] _ [tabspace] _ [ErrorExpr*] _ [WS] ') _ [WS]
	deconstruct X
		'info _ [': ?]
	by
		PrevFuncs [. Func]
end function

function getShellFunctions Func [FunctionCall]
	replace [FunctionCall*]
		PrevFuncs [FunctionCall*]
	deconstruct Func
		'$ '( X [ErrorName] _ [tabspace] _ [ErrorExpr*] _ [WS] ') _ [WS]
	deconstruct X
		'shell 
	by
		PrevFuncs [. Func]
end function

function findPaths_Assign A [Assignment]
	replace [Assignment*]
		PrevAssigns [Assignment*]
	where	
		A [containsPath_Assign]
	by
		PrevAssigns [. A]
end function

function containsPath_Assign
	match * [Assignment]
		A [Assignment]
	construct AssignsStr [stringlit]
		_ [unparse A]
	where	
		AssignsStr [grep "/"]
end function

function findPaths_Target T [Target]
	replace [Target*]
		PrevTargets [Target*]
	where	
		T [containsPath_Target]
	by
		PrevTargets [. T]
end function

function containsPath_Target
	match * [Target]
		T [Target]
	construct TargetStr [stringlit]
		_ [unparse T]
	where	
		TargetStr [grep "/"]
end function

function findPaths_Dependency D [Dependency]
	replace [Dependency*]
		PrevDependencies [Dependency*]
	where	
		D [containsPath_Dependency]
	by
		PrevDependencies [. D]
end function

function containsPath_Dependency
	match * [Dependency]
		D [Dependency]
	construct TargetStr [stringlit]
		_ [unparse D]
	where	
		TargetStr [grep "/"]
end function

function findPaths_Recipe C [RecipeExpr]
	replace [RecipeExpr*]
		PrevRecipeCommands [RecipeExpr*]
	where	
		C [containsPath_Recipe]
	by
		PrevRecipeCommands [. C]
end function

function containsPath_Recipe
	match * [RecipeExpr]
		C [RecipeExpr]
	construct CommandStr [stringlit]
		_ [unparse C]
	where	
		CommandStr [grep "/"]
end function

function findPaths_Id I [id]
	replace [id*]
		PrevIds [id*]
	where	
		I [containsPath_Id]
	by
		PrevIds [. I]
end function

function containsPath_Id
	match * [id]
		I [id]
	construct CommandStr [stringlit]
		_ [unparse I]
	where	
		CommandStr [grep "/"]
end function

rule countFunctionCalls
	replace $ [FunctionCall]
		Func [FunctionCall]
	skipping [FunctionCall]
	deconstruct Func
		'$ '( _ [FunctionName] _ [tabspace] Args [list FunctionArg+] _ [WS] ') _ [WS]
	
	construct RecursiveCalls [FunctionCall*]
		_ [^ Args]
	construct NRecursiveCalls [number]
		_ [length RecursiveCalls]
	deconstruct not NRecursiveCalls
		0
	
	import TotalEmbeddedFunctions [number]
	
	export TotalEmbeddedFunctions % [number]
		TotalEmbeddedFunctions [+ 1]
	
	by
		Func
end rule

rule countErrorFunctionCalls
	replace $ [FunctionCall]
		Func [FunctionCall]
	skipping [FunctionCall]
	deconstruct Func
		'$ '( _ [ErrorName] _ [tabspace] Args [ErrorExpr*] _ [WS] ') _ [WS]
	
	construct RecursiveCalls [FunctionCall*]
		_ [^ Args]
	construct NRecursiveCalls [number]
		_ [length RecursiveCalls]
	deconstruct not NRecursiveCalls
		0
	
	import TotalEmbeddedFunctions [number]
	
	export TotalEmbeddedFunctions % [number]
		TotalEmbeddedFunctions [+ 1]
	
	by
		Func
end rule

function getUniqueVars Ref [Evaluation]
	replace [Evaluation*]
		UniqueVars [Evaluation*]
	where not
		UniqueVars [containsVar Ref]
	by 
		UniqueVars [. Ref]
end function
 
rule containsVar Ref [Evaluation]
	match * [Evaluation]
		Ref
end rule

function getUniqueAssignmentVarRefs Assignment [Assignment]
	replace [Evaluation*]
		UniqueVars [Evaluation*]
	
	construct AssignmentRefs [Evaluation*]
		_ [^ Assignment]
	construct UniqueAssignmentRefs [Evaluation*]
		_ [getUniqueVars each AssignmentRefs]
	by 
		UniqueVars [. UniqueAssignmentRefs]
end function

function getUniqueTargetVarRefs Rule [Rule]
	replace [Evaluation*]
		UniqueVars [Evaluation*]
	construct Targets [Target*]
		_ [^ Rule]
	construct TargetRefs [Evaluation*]
		_ [^ Targets]
	construct UniqueTargetRefs [Evaluation*]
		_ [getUniqueVars each TargetRefs]
	by 
		UniqueVars [. UniqueTargetRefs]
end function

function getUniqueDependencyVarRefs Rule [Rule]
	replace [Evaluation*]
		UniqueVars [Evaluation*]
	construct Dependencies [Dependency*]
		_ [^ Rule]
	construct DependencyRefs [Evaluation*]
		_ [^ Dependencies]
	construct UniqueDependencyRefs [Evaluation*]
		_ [getUniqueVars each DependencyRefs]
	by 
		UniqueVars [. UniqueDependencyRefs]
end function

function getUniqueRecipeVarRefs Rule [Rule]
	replace [Evaluation*]
		UniqueVars [Evaluation*]
	
	construct Recipe [Recipe*]
		_ [^ Rule]
	construct RecipeRefs [Evaluation*]
		_ [^ Recipe]
	construct UniqueRecipeRefs [Evaluation*]
		_ [getUniqueVars each RecipeRefs]
	by 
		UniqueVars [. UniqueRecipeRefs]
end function


